%
% heuristik.pl
%
% Durfuehren der heuristischen Suche
%
% Dabei findet eine Orientierung an der Vorgenehnsweise eines 
% menschlichen Stundenplaners statt.  Dazu wird die 
% Sortierungsreihenfolge der Veranstatltungen geaendert
% (most_constrained zuerst).
%
% Eine Platzierung der Veranstaltungen wird mit den neu 
% sortierten Veranstaltungen versucht.
%
% Es ist moeglich, den Sonnabend und die Abendstunden aus 
% der Loesungsmenge zu entfernen.  Einen Entfernung aus der 
% Loesungsmenge findet nur statt, wenn eine Veranstaltung den
% Sonnabend bzw. den Abendstunden nicht explizit anfordert bzw.
% dieser Tag/die Abendstunden sich aus den auferlegten 
% Einschraenkungen nich ergibt.

/***********************************************************/
/* interneWissensbasis(-Wissensbasis)                      */
/***********************************************************/
/* Hier stehen alle Steuerungsdaten, um z.B. die           */
/* indizierte Liste zu erstellen.                          */
/***********************************************************/
interneWissensbasis('WBHeuristik.pl').

/***********************************************************/
/* verarbeitung(?Array)                                    */
/***********************************************************/
/* Ablaufsteuerung der heuristischen Suche,                */
/***********************************************************/
verarbeitung(Stundenplan) :-
	anzVeranst(AnzVeranst),
	erstelleZaehlvariablen(Raeume, Tage, Zeiten),
	interneWissensbasis(InterneWissensbasis),
	compile(InterneWissensbasis),
	erstelleIndexListe(Stundenplan, IndexListe),
	entferneSonnabend(Stundenplan, AnzVeranst),
	entferneAbendstunden(Stundenplan, AnzVeranst),
	labeling(Stundenplan, IndexListe, IndexListe, Raeume, Tage, Zeiten).

/***********************************************************/
/* erstelleIndexListe(?Stundenplan, -Ergebnis)             */
/***********************************************************/
/* Die eingegebenen Veranstaltungen werden losgeloest von  */
/* ihrer Eingabereihenfolge betrachtet.                    */
/* Es besteht die Moeglichkeit, die Sortierung der Werte   */
/* zu beeinflussen.  Dazu kann bei der Addition der        */
/* Ergebniswerte in addiereWerte die Wertigkeiten der      */
/* einzelnen Zahlen durch Multiplikation, Division,        */
/* Addition und Subtraktion beeinflusst werden.  Die       */
/* korrekte Einstellung wird mit Trial-and-Error gefunden. */
/***********************************************************/
erstelleIndexListe(Stundenplan, Ergebnis) :-
	anzVeranst(AnzVeranst),
	findall((ID, Teilnehmer), veranstaltung(ID, _, Teilnehmer, _, _), TeilnehmerListe),
	findall((ID, Anzahl), veranstaltung(ID, _, _, _, Anzahl), AnzahlListe),
	% erstmal alle Veranstaltungen finden, 
	% die fest bzw. stark eingeschraenkt sind ...
	% so ist gesichert, dass diese immer zuerst platziert werden,
	% da sie mit 3 000 000\2 000 000\1 000 000 in die Wertung eingehen
	% 3 000 000
	berechneFesteVeranstaltungen(Stundenplan, AnzVeranst, FesteVeranstaltungenListe),
	berechneTeilnehmerVerfuegbarkeiten(TeilnehmerVerfuegbarkeitenListe),
	% 2 000 000
	berechneGrosseTeilnehmerGruppen(TeilnehmerListe, TeilnehmerGruppenListe),
	berechneDreiUndVierfachstunden(DreiUndVierfachStundenListe),
	% 1 000 000
	berechneDoppelstunden(DoppelstundenListe),
	append(FesteVeranstaltungenListe, TeilnehmerVerfuegbarkeitenListe, Liste1),
	append(Liste1, TeilnehmerGruppenListe, Liste2),
	append(Liste2, DreiUndVierfachStundenListe, Liste3),
	append(Liste3, DoppelstundenListe, Liste4),
	sort(2, >, Liste4, HoechsteEinschraenkungenListe),
	schreibeAufVeranstaltungenUm2(HoechsteEinschraenkungenListe, AnzVeranst, HEGesamtListe),
	% dann den Rest ...
	berechneMostConstraint(Stundenplan, AnzVeranst, MostConstraintListe),
	berechneTeilnehmerGruppenAnzahl(TeilnehmerListe, TeilnehmerGruppenListe2),
	berechneTeilnehmerAnzahl(AnzahlListe, TeilnehmerAnzahlListe),
	berechneTeilnehmerEinschraenkungen(EinschraenkungsListe),
	sort(2, >, TeilnehmerGruppenListe2, TeilnehmerGruppenListeSort),
	sort(2, >, MostConstraintListe, MostConstraintListeSort),
	sort(2, >, TeilnehmerAnzahlListe, TeilnehmerAnzahlListeSort),
	sort(2, >, EinschraenkungsListe, EinschraenkungsListeSort),
	addiereWerte(TeilnehmerGruppenListeSort, MostConstraintListeSort, TeilnehmerAnzahlListeSort, EinschraenkungsListeSort, HEGesamtListe, KombiListe),
	sort(0, >, KombiListe, KombiListeSort),
	bestimmeSortiertenIndex(KombiListeSort, Ergebnis).

/***********************************************************/
/* berechneFesteEinschraenkungen(?Stundenplan, +Index,     */
/*	-Ergebnis)                                         */
/***********************************************************/
/* Herausfinden, welche Veranstaltungen keinen             */
/* Platzierungsspielraum mehr zur Verfuegung stellen.      */
/***********************************************************/
berechneFesteVeranstaltungen(_, 0, []).
berechneFesteVeranstaltungen(Stundenplan, Index, [Kopf|Rest]) :-
	subscript(Stundenplan, [Index, 1], Raum),
	subscript(Stundenplan, [Index, 2], Tag),
	subscript(Stundenplan, [Index, 3], Zeit),
	delayed_goals_number(Raum, RaumC),
	delayed_goals_number(Tag, TagC),
	delayed_goals_number(Zeit, ZeitC),
	Ergebnis is (RaumC + TagC + ZeitC),
	Ergebnis >= 3000000,
	append([Ergebnis], [Index], Kopf),
	IndexNeu is Index - 1,
	berechneFesteVeranstaltungen(Stundenplan, IndexNeu, Rest).
berechneFesteVeranstaltungen(Stundenplan, Index, Ergebnis) :-
	IndexNeu is Index - 1,
	berechneFesteVeranstaltungen(Stundenplan, IndexNeu, Ergebnis).

/***********************************************************/
/* berechneTeilnehmerVerfuegbarkeiten(-Ergebnis)           */
/***********************************************************/
/* Bestimmen, welche Teilnehmer nur so viel                */
/* Platzierungsmoeglichkeiten bieten wie sie               */
/* Veranstaltungsverpflichtungen haben.                    */
/***********************************************************/
berechneTeilnehmerVerfuegbarkeiten(Ergebnis) :-
	berechneVerfuegbarkeitEinesTeilnehmers(Liste),
	schreibeRelevanteVeranstaltungen(Liste, 3000000, Ergebnis).
	
/***********************************************************/
/* schreibeRelevanteVeranstaltungen(+Liste, +Wert,         */
/*	-Ergebnis)                                         */
/***********************************************************/
/* Fuegt die Veranstaltungen in die Liste ein, deren       */
/* Teilnehmer genau so viele freie Zeiteinheiten wie       */
/* Veranstaltungsverpflichtungen besitzen.                 */
/* Dies sind die Teilnehmer, bei denen Zahl den Wert 0     */
/* besitzt.                                                */
/***********************************************************/
schreibeRelevanteVeranstaltungen([], _, []).
schreibeRelevanteVeranstaltungen([Zahl, ID|Rest], Wert, [ErgebnisKopf|ErgebnisRest]) :-
	Zahl == 0,
	append([Wert], [ID], ErgebnisKopf),
	schreibeRelevanteVeranstaltungen(Rest, Wert, ErgebnisRest).
schreibeRelevanteVeranstaltungen([Zahl, _|Rest], Wert, Ergebnis) :-
	Zahl =\= 0,
	schreibeRelevanteVeranstaltungen(Rest, Wert, Ergebnis).

/***********************************************************/
/* berechneGrosseTeilnehmerGruppen(+Liste, -Ergebnis)      */
/***********************************************************/
/* Auffinden von  Veranstaltungen, die fuer eine grosse    */
/* Anzahl von Teilnehmergruppen gehalten werden.           */
/***********************************************************/
berechneGrosseTeilnehmerGruppen([], []).
berechneGrosseTeilnehmerGruppen([(ID, Teilnehmer)|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	length(Teilnehmer, N),
	N > 6,
	append([2000000, ID], [], ErgebnisKopf),
	berechneGrosseTeilnehmerGruppen(Rest, ErgebnisRest).
berechneGrosseTeilnehmerGruppen([(_, _)|Rest], Ergebnis) :-
	berechneGrosseTeilnehmerGruppen(Rest, Ergebnis).

/***********************************************************/
/* berechneDreiUndVierfachstunden(-Ergebnis)               */
/***********************************************************/
/* Berechnen der Drei- und Vierfachstunden.  Diese werden  */
/* mit einer Prioritaet versehen, um so dafr zu sorgen,   */
/* dass sie fruehzeitig platziert werden.                  */
/***********************************************************/
berechneDreiUndVierfachstunden(Ergebnis) :-
	findeDreifachstunden(DreifachstundenListe),
	findeVierfachstunden(VierfachstundenListe),
	append(DreifachstundenListe, VierfachstundenListe, Liste),
	wandleVeranstaltungenInZahlen(Liste, ZahlenListe),
	flatten(ZahlenListe, ZahlenListeFlat),
	schreibeRelevanteVeranstaltungen(ZahlenListeFlat, 2000000, Ergebnis).

/***********************************************************/
/* berechneDoppelstunden(-Ergebnis)                        */
/***********************************************************/
/* Arbeit analog zu berechneDreiUndVierfachstunden/1, nur  */
/* die Prioritaet ist eine andere.                         */
/***********************************************************/
berechneDoppelstunden(Ergebnis) :-
	findeDoppelstunden(Liste),
	wandleVeranstaltungenInZahlen(Liste, ZahlenListe),
	flatten(ZahlenListe, ZahlenListeFlat),
	schreibeRelevanteVeranstaltungen(ZahlenListeFlat, 1000000, Ergebnis).

/***********************************************************/
/* berechneTeilnehmerGruppenAnzahl(+Teilnehmer, -Ergebnis) */
/***********************************************************/
/* Berechnen der Anzahl der Teilnehmergruppen einer        */
/* Veranstaltung.                                          */
/***********************************************************/
berechneTeilnehmerGruppenAnzahl([], []).
berechneTeilnehmerGruppenAnzahl([(ID, Teilnehmer)|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	length(Teilnehmer, N),
	teilnehmerGruppen(Wert),
	NNeu is N * Wert,
	append([NNeu, ID], [], ErgebnisKopf),
	berechneTeilnehmerGruppenAnzahl(Rest, ErgebnisRest).

/***********************************************************/
/* berechneMostConstraint(?Stundenplan, +Index, -Ergebnis) */
/***********************************************************/
/* Herausfinden, welche Veranstaltungen mit den meisten    */
/* Einschraenkungen belegt ist.                            */
/* "delayed_goals_in_number" gibt 1 000 000 zurueck, wenn  */
/* ein Wert fest gesetzt ist.  Ansonsten wird die Anzahl   */
/* der auf diesem Wert liegenden Einschraenkungen          */
/* zurueckgegeben.                                         */
/* Ist eine Veranstaltung auf einen Tag, einen Raum und    */
/* eine Zeit eingeschraenkt, bekommt sie den Wert          */
/* 1000000 + 1000000 + 1000000 = 3000000 zugewiesen.       */
/* Der Ergebniswert (max. 3 000 000) wird noch normiert,   */
/* so dass er in das Gesamtzahlenschema hineinpasst.       */
/***********************************************************/
berechneMostConstraint(_, 0, []).
berechneMostConstraint(Stundenplan, Index, [Kopf|Rest]) :-
	subscript(Stundenplan, [Index, 1], Raum),
	subscript(Stundenplan, [Index, 2], Tag),
	subscript(Stundenplan, [Index, 3], Zeit),
	delayed_goals_number(Raum, RaumC),
	delayed_goals_number(Tag, TagC),
	delayed_goals_number(Zeit, ZeitC),
	Ergebnis is (RaumC + TagC + ZeitC),
	Ergebnis > 1000000,
	% der default-Wert ist 1000, so dass der max-Wert, der in 
	% die Berechnung eingeht, bei 3000 liegt.
	mostConstraint(Wert),
	ErgebnisNeu is Ergebnis / Wert,
	append([ErgebnisNeu], [Index], Kopf),
	IndexNeu is Index - 1,
	berechneMostConstraint(Stundenplan, IndexNeu, Rest).
berechneMostConstraint(Stundenplan, Index, [Kopf|Rest]) :-
	subscript(Stundenplan, [Index, 1], Raum),
	subscript(Stundenplan, [Index, 2], Tag),
	subscript(Stundenplan, [Index, 3], Zeit),
	delayed_goals_number(Raum, RaumC),
	delayed_goals_number(Tag, TagC),
	delayed_goals_number(Zeit, ZeitC),
	% Ergebnis zu Fliesskommazahl machen
	Ergebnis is (RaumC + TagC + ZeitC) / 1,
	append([Ergebnis], [Index], Kopf),
	IndexNeu is Index - 1,
	berechneMostConstraint(Stundenplan, IndexNeu, Rest).

/***********************************************************/
/* berechneTeilnehmerAnzahl(+Liste, - Ergebnis)            */
/***********************************************************/
/* Wenn ein Raum fest vorgegeben ist, wird Anzahl auf      */
/* den vorgegebenen Wert teilnehmerAnzahl(_) gesetzt.      */
/* Ist kein Raum gegeben, wird der Originalwert genommen.  */
/***********************************************************/
berechneTeilnehmerAnzahl([], []).
berechneTeilnehmerAnzahl([(ID, 0)|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	% Auswahl auf ein oder mehrere Raeume eingeschraenkt
	teilnehmerAnzahl(Anzahl),
	append([Anzahl, ID], [], ErgebnisKopf),
	berechneTeilnehmerAnzahl(Rest, ErgebnisRest).
berechneTeilnehmerAnzahl([(ID, Anzahl)|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	% keine Einschraenkung bei der Raumwahl getroffen
	append([Anzahl, ID], [], ErgebnisKopf),
	berechneTeilnehmerAnzahl(Rest, ErgebnisRest).

/***********************************************************/
/* berechneTeilnehmerEinschraenkungen(-Ergebnis)           */
/***********************************************************/
/* Stelle fest, welche Teilnehmer in der Wissensbasis so   */
/* viele Freiraeume gegeben haben wie sie Veranstaltungen  */
/* besuchen muessen.                                       */
/* Diese Veranstaltungen muessen zu Anfang platziert       */
/* werden, so dass die Zeitraeume dieser Veranstaltungen   */
/* noch nicht anderweitig belegt sind.                     */
/***********************************************************/
berechneTeilnehmerEinschraenkungen(Ergebnis) :-
	anzVeranst(AnzVeranst),
	berechneVerfuegbarkeitEinesTeilnehmers(Liste),
	schreibeAufVeranstaltungenUm(Liste, ErgebnisListe),
	sort(2, >, ErgebnisListe, ErgebnisListeSort),
	schreibeAufVeranstaltungenUm2(ErgebnisListeSort, AnzVeranst, Ergebnis).

/***********************************************************/
/* schreibeAufVeranstaltungenUm(+Liste, -Ergebnis)         */
/***********************************************************/
/* Schreiben der Werte fuer gefundene Veranstaltungen.     */
/* Dabei werden niedrige Werte in hohe umgewandelt.        */
/***********************************************************/
schreibeAufVeranstaltungenUm([], []).
schreibeAufVeranstaltungenUm([Zahl, ID|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	teilnehmerEinschraenkung(Wert),
	% Umwandlung von niedrigen Werten in 
	% hohe Werte anhand einer Konstante
	ZahlNeu is Wert - Zahl,
	append([ZahlNeu], [ID], ErgebnisKopf),
	schreibeAufVeranstaltungenUm(Rest, ErgebnisRest).

/***********************************************************/
/* schreibeAufVeranstaltungenUm2(+Liste, +Index,           */
/*	-Ergebnis)                                         */ 
/***********************************************************/
/* Schreiben der Werte fuer alle(!) Veranstaltungen.       */
/* Neue Indizes werden mit 0 belegt.                       */
/***********************************************************/
schreibeAufVeranstaltungenUm2([], 0, []).
schreibeAufVeranstaltungenUm2([], Index, [ErgebnisKopf|ErgebnisRest]) :-
	% damit am Ende keine Werte vergessen werden ...
	Index =\= 0,
	append([0], [Index], ErgebnisKopf),
	IndexNeu is Index - 1,
	schreibeAufVeranstaltungenUm2([], IndexNeu, ErgebnisRest).
schreibeAufVeranstaltungenUm2([[Zahl, ID]|Rest], Index, [ErgebnisKopf|ErgebnisRest]) :-
	Index == ID,
	append([Zahl], [ID], ErgebnisKopf),
	IndexNeu is Index - 1,
	schreibeAufVeranstaltungenUm2(Rest, IndexNeu, ErgebnisRest).
schreibeAufVeranstaltungenUm2([[Zahl, ID]|Rest], Index, [ErgebnisKopf|ErgebnisRest]) :-
	% wenn keine weiteren Einschraenkungen bei 
	% teilnehmerNichtAmUm/Mehrfachstunden gesetzt
	Index =\= ID,
	append([0], [Index], ErgebnisKopf),
	IndexNeu is Index - 1,
	schreibeAufVeranstaltungenUm2([[Zahl, ID]|Rest], IndexNeu, ErgebnisRest).

/***********************************************************/
/* berechneMehrfachstunden(-Ergebnis)                      */ 
/***********************************************************/
/* Gewichtung fuer Drei- und Vierfachstunden berechnen     */
/***********************************************************/
berechneMehrfachstunden(MehrfachstundenListe) :-
	anzVeranst(AnzVeranst),
	findeDreifachstunden(DreifachstundenListe),
	findeVierfachstunden(VierfachstundenListe),
	append(DreifachstundenListe, VierfachstundenListe, Liste),
	wandleVeranstaltungenInZahlen(Liste, ZahlenListe),
	sort(2, >,  ZahlenListe, ZahlenListeSort),
	schreibeAufVeranstaltungenUm2(ZahlenListeSort, AnzVeranst, MehrfachstundenListe).

/***********************************************************/
/* addiereWerte(+Liste1, +Liste2, +Liste3, +Liste4,        */
/*	+Liste5, -Ergebnis)                                */
/***********************************************************/
/* Zusammenzaehlen der Werte auf deren Basis die           */
/* indizierte Liste erstellt wird.                         */
/***********************************************************/
addiereWerte([], [], [], [], [], []).
addiereWerte([[Zahl1, ID]|Rest1], [[Zahl2, ID]|Rest2], [[Zahl3, ID]|Rest3], [[Zahl4, ID]|Rest4], [[Zahl5, ID]|Rest5], [ErgebnisKopf|ErgebnisRest]) :-
	Zahl is Zahl1 + Zahl2 + Zahl3 + Zahl4 + Zahl5,
	append([Zahl], [ID], ErgebnisKopf),
	addiereWerte(Rest1, Rest2, Rest3, Rest4, Rest5, ErgebnisRest).

/***********************************************************/
/* bestimmeSortiertenIndex(+Liste, -Ergebnis)              */
/***********************************************************/
/* Sorgt dafuer, dass das letze Element - der Index - aus  */
/* der Liste extrahiert wird.                              */
/***********************************************************/
bestimmeSortiertenIndex([], []).
bestimmeSortiertenIndex([Kopf|Rest], [ErgebnisKopf|ErgebnisRest]):-
	letztes(ErgebnisKopf, Kopf),
	bestimmeSortiertenIndex(Rest, ErgebnisRest).

/***********************************************************/
/* erstelleZaehlvariablen(-Raeume, -Tage, -Zeiten)         */
/***********************************************************/
/* Hier werden die Arrays erstellt, die bei der Erstellung */
/* der Reihenfolge der Domainvariablen bei der Platzierung */
/* verwendet werden.                                       */
/***********************************************************/
erstelleZaehlvariablen(Raeume, Tage, Zeiten) :-
	anzRaeume(AnzRaeume),
	anzTage(AnzTage),
	anzZeiten(AnzZeiten),
	functor(Raeume, raeume, AnzRaeume),
	functor(Tage, tage, AnzTage),
	functor(Zeiten, zeiten, AnzZeiten),
	initZaehlvariablen(Raeume),
	initZaehlvariablen(Tage),
	initZaehlvariablen(Zeiten).

/***********************************************************/
/* initZaehlvariablen(?Array)                              */
/***********************************************************/
/* Initialisieren der Arraywerte mit 0.                    */
/***********************************************************/
initZaehlvariablen(Array) :-
	(
	    foreacharg(X, Array) 
	do 
	    X is 0
	).

/***********************************************************/
/* entferneSonnabend(?Stundenplan, +Index)                 */
/***********************************************************/
/* Der Sonnabend wird aus dem Suchraum entfernt, dabei     */
/* werden nur Veranstaltungen bercksichtigt, die den      */
/* Sonnabend nicht explizit anfordern.                     */
/* Auch findet eine Ueberpruefung nur statt, wenn mehr     */
/* als 5 Tage die Woche unterrichtet werden.               */
/***********************************************************/
entferneSonnabend(Stundenplan, Index) :-
	findall(Tage, tag(Tage, _, _), TagListe),
	length(TagListe, M),
	% mehr Tage als eine 5-Tage-Woche ?
	% -> die zusaetzlichen Tage nur bei Anforderung nutzen
	M > 5,
	AnzTage is 5, 
	entferneSonnabend(Stundenplan, Index, AnzTage).

/***********************************************************/
/* entferneSonnabend(?Stundenplan, +ZaehlIndex,            */
/*	+GesamtIndex)                                      */
/***********************************************************/
entferneSonnabend(_, 0, _).
entferneSonnabend(Stundenplan, Index, AnzTage) :-
	subscript(Stundenplan, [Index, 2], Tag),
	nonvar(Tag),
	IndexNeu is Index - 1,
	entferneSonnabend(Stundenplan, IndexNeu, AnzTage).
entferneSonnabend(Stundenplan, Index, AnzTage) :-
	subscript(Stundenplan, [Index, 2], Tag),
	% trotz der Zuweisung von 1..AnzTage an Tag, 
	% werden nur die Tage zugewiesen, die sich 
	% wirklich noch im Suchraum befinden.
	Tag :: 1..AnzTage, 
	IndexNeu is Index - 1,
	entferneSonnabend(Stundenplan, IndexNeu, AnzTage).

/***********************************************************/
/* entferneAbendstunden(?Stundenplan, +Index)              */
/***********************************************************/
/* Die Abendstunden werden aus dem Suchraum entfernt,      */
/* dabei werden nur Veranstaltungen bercksichtigt, die    */
/* die Abendstunden nicht explizit anfordern.              */
/* Auch findet eine Ueberpruefung nur statt, wenn mehr     */
/* als 7 Stunden pro Tag unterrichtet werden.              */
/***********************************************************/
entferneAbendstunden(Stundenplan, Index) :-
	findall(Zeiten, zeit(Zeiten, _, _), ZeitListe),
	length(ZeitListe, M),
	% mehr als 7 Vorlesungszeiten?
	% -> die zusaetzlichen Zeiten nur nach Anforderung nutzen
	M > 7,
	AnzZeiten is 7,
	entferneAbendstunden(Stundenplan, Index, AnzZeiten).

/***********************************************************/
/* entferneAbendstunden(?Stundenplan, +ZaehlIndex,         */
/*	+GesamtIndex)                                      */
/***********************************************************/
entferneAbendstunden(_, 0, _).
entferneAbendstunden(Stundenplan, Index, AnzZeiten) :-
	subscript(Stundenplan, [Index, 3], Zeit),
	nonvar(Zeit),
	IndexNeu is Index - 1,
	entferneAbendstunden(Stundenplan, IndexNeu, AnzZeiten).
entferneAbendstunden(Stundenplan, Index, AnzZeiten) :-
	subscript(Stundenplan, [Index, 3], Zeit),
	% trotz der Zuweisung von 1..AnzZeiten an Zeit, 
	% werden nur die Zeiten zugewiesen, die sich 
	% wirklich noch im Suchraum befinden.
	Zeit :: 1..AnzZeiten, 
	IndexNeu is Index - 1,
	entferneAbendstunden(Stundenplan, IndexNeu, AnzZeiten).

/***********************************************************/
/* labeling(?Array, +Index, +Raeume, +Tag, +Zeiten)        */
/***********************************************************/
/* Belegen aller Variablen mit Werten ihrer Domain         */
/* Die eigentliche Platzierung findet in                   */
/* sucheGuenstigstePlatzierung statt.  Die Buchfuehrung    */
/* erstellt die Arrays, die im naechsten Suchlauf benutzt  */
/* werden.  Die Berechnung wird in jedem Durchlauf von     */
/* neuem durchgefuehrt, damit die Arrays vom Backtracking  */
/* nicht beeinflusst werden.  Die Arrays werden vom        */
/* Backtracking nicht beruecksichtigt.  Deshalb ist eine   */
/* Neuberechnung erforderlich.                             */  
/***********************************************************/
labeling(_, [], _, _, _, _).
labeling(Stundenplan, [Index|Rest], IndexListe, Raeume, Tage, Zeiten) :-
	% Verwaltung
	veranstaltung(Index, Name, _, _, _),
	% Debug-Ausgabe - ueberpruefen, wie der Suchablauf fortschreitet
	write('Versuche '),
	write(Name),
	writeln(' zu platzieren.'),
	write('Es verbleiben noch '),
	length(Rest, N),
	write(N),
	writeln(' Veranstaltungen'),
	writeln(''),
	subscript(Stundenplan, [Index, 1], Raum),
	subscript(Stundenplan, [Index, 2], Tag),
	subscript(Stundenplan, [Index, 3], Zeit),
	sucheGuenstigstePlatzierung(Raum, Tag, Zeit, Raeume, Tage, Zeiten),
	% Buchfuehrung
	anzRaeume(AnzRaeume),
	buchfuehren(Stundenplan, Raeume, Index, IndexListe, AnzRaeume, 1),
	anzTage(AnzTage),
	buchfuehren(Stundenplan, Tage, Index, IndexListe, AnzTage, 2),
	anzZeiten(AnzZeiten),
	buchfuehren(Stundenplan, Zeiten, Index, IndexListe, AnzZeiten, 3),
	labeling(Stundenplan, Rest, IndexListe, Raeume, Tage, Zeiten).

/***********************************************************/
/* buchfuehren(+Stundenplan, ?Array, +Index, +IndexListe,  */
/*	+Anzahl, +Element)                                 */
/***********************************************************/
/* Dieses Praedikat wird genutzt, um die Durchfuehrung der */
/* Buchfuehrung zu steuern.                                */
/* Array - zu bearbeitendes Array                          */
/* Index - aktueller Index                                 */
/* IndexListe - Liste ALLER Indizes                        */
/* Anzahl - Groesse des Array                              */
/* Element - zweite zahl, die beim subscript benoetigt     */
/*	wird, um zw. Raum, Tag und Zeit zu unterscheiden   */
/***********************************************************/
buchfuehren(Stundenplan, Array, Index, IndexListe, Anzahl, Element) :-
	% array initialisieren mit 0
	(
	    count(I, 1, Anzahl), 
	    param(Array) 
	do 
	    setarg(I, Array, 0)
	),
	buchfuehren1(Stundenplan, Array, Index, IndexListe, Element).

/***********************************************************/
/* buchfuehren1(+Stundenplan, ?Array, +Index, +Liste,      */
/*	+Element)                                          */
/***********************************************************/
/* Bei der Buchfuehrung werden nur die Veranstaltungen     */
/* beruecksichtigt, die schon instanziiert sind.           */
/***********************************************************/
buchfuehren1(Stundenplan, Array, Index, [Kopf|Rest], Element) :-
	Index =\= Kopf,
	subscript(Stundenplan, [Kopf, Element], Wert),
	arg(Wert, Array, AlterWert),
	NeuerWert is AlterWert + 1,
	setarg(Wert, Array, NeuerWert),
	buchfuehren1(Stundenplan, Array, Index, Rest, Element).
buchfuehren1(Stundenplan, Array, Index, [Index|_], Element) :-
	subscript(Stundenplan, [Index, Element], Wert),
	arg(Wert, Array, AlterWert),
	NeuerWert is AlterWert + 1,
	setarg(Wert, Array, NeuerWert).	

/***********************************************************/
/* sucheGuenstigstePlatzierung(?Raum, ?Tag, ?Zeit,         */
/*	+Raeume, +Tage, +Zeiten)                           */
/***********************************************************/
/* Hier findet das eigentliche Platzieren der              */
/* Veranstaltungen im Plan statt.  Die einzelnen Arrays    */
/* dienen dazu, die Sortierung der im Wertebereich         */
/* verbleibenden Elemente zu beeinflussen.  Die ermittlte  */
/* Reihenfolge wird beim Zuweisen durchlaufen.             */
/***********************************************************/
sucheGuenstigstePlatzierung(Raum, Tag, Zeit, Raeume, Tage, Zeiten) :-
	dom(Raum, RaumListe),
	erstelleSortierteListe(RaumListe, Raeume, SortierteRaumListe),
	mein_indomain(Raum, SortierteRaumListe),
	dom(Tag, TagListe),
	erstelleSortierteListe(TagListe, Tage, SortierteTagListe),
	mein_indomain(Tag, SortierteTagListe),
	dom(Zeit, ZeitListe),
	erstelleSortierteListe(ZeitListe, Zeiten, SortierteZeitListe),
	mein_indomain(Zeit, SortierteZeitListe).

/***********************************************************/
/* addiereListen(+SortierteListe, +Liste, -Ergebnis)       */
/***********************************************************/
/* Zusammenfuegen von zwei Listen durch Addition der Werte */
/***********************************************************/
addiereListen(Ergebnis, [], Ergebnis).
addiereListen(SortierteTagListe, [Kopf|Rest], Ergebnis) :-
	addiereListen1(SortierteTagListe, Kopf, AdditionsListe),
	addiereListen(AdditionsListe, Rest, Ergebnis).

/***********************************************************/
/* addiereListen1(+Liste1, +Liste2, -Ergebnis)             */
/***********************************************************/
addiereListen1([], [], []).
addiereListen1([Kopf1|Rest1], [Kopf2|Rest2], [ErgebnisKopf|ErgebnisRest]) :-
	ErgebnisKopf is Kopf1 + Kopf2,
	addiereListen1(Rest1, Rest2, ErgebnisRest).

/***********************************************************/
/* erstelleSortierteListe(+Liste, +Array,                  */
/*	-KriteriumsListe)                                  */
/***********************************************************/
/* Erstellt eine Liste, die sortiert nach den              */
/* Haeufigkeiten der Vorkommnisse ist.                     */
/***********************************************************/
erstelleSortierteListe([], _, []).
erstelleSortierteListe([Kopf|Rest], Array, [Kriterium|KRest]) :-
	arg(Kopf, Array, Wert),
	Kriterium = Wert,
	erstelleSortierteListe(Rest, Array, KRest).

/***********************************************************/
/* mein_indomain(Variable, [Liste])                        */
/***********************************************************/
/* Belegen einer Variablen mit einem Wert ihrer Domain     */
/***********************************************************/
mein_indomain(DomainVar, _) :-
	nonvar(DomainVar).
mein_indomain(DomainVar, KriteriumsListe) :-
	var(DomainVar),
	dom(DomainVar, WertebereichsListe),
	sortiereWerte(WertebereichsListe, KriteriumsListe, SortierteWertebereichsListe),
	member(DomainVar, SortierteWertebereichsListe).

/***********************************************************/
/* sortiereWerte(+Liste1, +Liste2, -Ergebnis)              */
/***********************************************************/
/* Sortieren der Liste1 unter Zurhilfenahme der Liste2.    */
/***********************************************************/
sortiereWerte(Liste1, Liste2, Ergebnis) :-
	erstelleKombiListe(Liste2, Liste1, KombiListe),
	sort(KombiListe, SortierteKombiListe),
	erstelleWerteListe(SortierteKombiListe, Ergebnis).

/***********************************************************/
/* erstelleKombiListe(+Liste1, +Liste2, -Ergebnis)         */
/***********************************************************/
/* Kombination der beiden Listen.                          */
/***********************************************************/
erstelleKombiListe([], [], []).
erstelleKombiListe([Kopf1|Rest1], [Kopf2|Rest2], [ErgebnisKopf|ErgebnisRest]) :-
	append([Kopf1], [Kopf2], ErgebnisKopf),
	erstelleKombiListe(Rest1, Rest2, ErgebnisRest).

/***********************************************************/
/* erstelleWerteListe(+Liste, -Ergebnis)                   */
/***********************************************************/
/* Herausfiltern der relevanten Werte aus der Liste.       */
/***********************************************************/
erstelleWerteListe([], []).
erstelleWerteListe([Kopf|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	letztes(ErgebnisKopf, Kopf),
	erstelleWerteListe(Rest, ErgebnisRest).

/***********************************************************/
/* berechneVerfuegbarkeitEinesTeilnehmers(-Ergebnis)       */
/***********************************************************/
/* Berechnen der zur Verfuegung stehenden Zeiteinheiten    */
/* eines Teilnehmers.  Gleichzeitig geschieht ein Abgleich */
/* mit den Veranstaltungsverpflichtungen eines             */
/* Teilnehmers.  Das Ergebnis ist 0, wenn ein Teilnehmer   */
/* genau so viele Veranstatlungsverpflichtungen besitzt    */
/* wie er freie Zeiteinheiten besitzt.                     */
/***********************************************************/
berechneVerfuegbarkeitEinesTeilnehmers(Ergebnis) :-
	% Verwaltung
	anzTage(AnzTage),
	anzZeiten(AnzZeiten),
	findall(Tage, tag(Tage, _), TagListe),
	% Intelligenz
	% alle benutzten Praedikate finden sich in 
	% 'allgemeineFunktionaltitaeten.pl'
	findall(Teilnehmer, teilnehmerNichtAmUm(Teilnehmer, _, _), TeilnehmerListe),
	sort(TeilnehmerListe, TeilnehmerListeSort),
	findeMehrfachAufgefuehrte(TeilnehmerListeSort, MehrfachListe),
	erweitereListeUmTage(MehrfachListe, TagListe, NeueTagListe),
	schreibeTage(MehrfachListe, NeueTagListe, NeueMehrfachListe),
	flatten(NeueMehrfachListe, NeueMehrfachListeFlat),
	findall((Teilnehmer, Am, Um), (teilnehmerNichtAmUm(Teilnehmer, Am, Um), nonmember((Teilnehmer, _, _), NeueMehrfachListeFlat)), NeueTeilnehmerListe),
	append(NeueTeilnehmerListe, NeueMehrfachListeFlat, VollstaendigeTeilnehmerListe),
	sort(VollstaendigeTeilnehmerListe, VollstaendigeTeilnehmerListeSort),
	berechneZurVerfuegungStehendeZeiteinheiten(VollstaendigeTeilnehmerListeSort, AnzTage, AnzZeiten, ZeiteinheitenListe),
	flatten(ZeiteinheitenListe, ZeiteinheitenListeFlat),
	fuehreErgebnisseZusammen(ZeiteinheitenListeFlat, ZeiteinheitenListeKomprimiert),
	flatten(ZeiteinheitenListeKomprimiert, ZeiteinheitenListeKomprimiertFlat),
	AnzZeitenUndTage is AnzTage * AnzZeiten,
	% Intelligenz - Ende
	passeZahlenAn(ZeiteinheitenListeKomprimiertFlat, AnzZeitenUndTage, AngepassteListe),
	findeAnzahlDerVeranstaltungen(AngepassteListe, SpielraumListe),
	rechneAufVeranstaltungenUm(SpielraumListe, ErgebnisListe),
	flatten(ErgebnisListe, Ergebnis).

/***********************************************************/
/* passeZahlenAn(+Liste, +Zahl, -Ergebnis)                 */
/***********************************************************/
/* Umrechnung des Ergebnisses von freien Zeiteinheiten auf */
/* zur Verfuegung stehende Zeiteinheiten.                  */
/***********************************************************/
passeZahlenAn([], _, []).
passeZahlenAn([(Name, Stunden)|Rest], Zahl, [ErgebnisKopf|ErgebnisRest]) :-
	StundenNeu is Zahl - Stunden,
	append([Name], [StundenNeu], ErgebnisKopf),
	passeZahlenAn(Rest, Zahl, ErgebnisRest).

/***********************************************************/
/* findeAnzahlDerVeranstaltungen(+VeransListe, -Ergebnis)  */
/***********************************************************/
/* Gibt den Spielraum der einzelnen Teilnehmer aus.  Ein   */
/* Teilnehmer mit 8 freien Zeiteinheiten und genau 8       */
/* Veranstaltungen bekommt den Wert 0 zugeteilt.           */
/***********************************************************/
findeAnzahlDerVeranstaltungen([], []).
findeAnzahlDerVeranstaltungen([Kopf|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	findeAnzahlDerVeranstaltungen1(Kopf, ErgebnisKopf),
	findeAnzahlDerVeranstaltungen(Rest, ErgebnisRest).

/***********************************************************/
/* findeAnzahlDerVeranstaltungen1(+VeransListe, -Ergebnis) */
/***********************************************************/
findeAnzahlDerVeranstaltungen1([Teilnehmer, Anzahl], Ergebnis) :-
	findall(Name, (veranstaltung(_, Name, TeilnehmerListe, _, _), member(Teilnehmer, TeilnehmerListe)), Liste),
	length(Liste, N),
	AnzahlNeu is Anzahl - N,
	append([Teilnehmer], [AnzahlNeu], Ergebnis).

/***********************************************************/
/* rechneAufVeranstaltungenUm(+Liste, -Ergebnis)           */
/***********************************************************/
/* Umrechnen des bis dato gefundenen teilnehmerbezogenen   */
/* Ergebnisses auf Veranstaltungen.                        */
/***********************************************************/
rechneAufVeranstaltungenUm([], []).
rechneAufVeranstaltungenUm([Kopf|Rest], [ErgebnisKopf|ErgebnisRest]) :-
	rechneAufVeranstaltungenUm1(Kopf, ErgebnisKopf),
	rechneAufVeranstaltungenUm(Rest, ErgebnisRest).

/***********************************************************/
/* rechneAufVeranstaltungenUm1(+Listenkopf, -Ergebnis)     */
/***********************************************************/
rechneAufVeranstaltungenUm1([Teilnehmer, Zahl], Ergebnis) :-
	findall(ID, (veranstaltung(ID, _, TeilnehmerListe, _, _), member(Teilnehmer, TeilnehmerListe)), Liste),
	rechneAufVeranstaltungenUm2(Liste, Zahl, Ergebnis).

/***********************************************************/
/* rechneAufVeranstaltungenUm2(+Liste, +Prio, -Ergebnis)   */
/***********************************************************/
rechneAufVeranstaltungenUm2([], _, []).
rechneAufVeranstaltungenUm2([Kopf|Rest], Zahl, [ErgebnisKopf|ErgebnisRest]) :-
	append([Zahl], [Kopf], ErgebnisKopf),
	rechneAufVeranstaltungenUm2(Rest, Zahl, ErgebnisRest).
